LITEMATIC_VERSION = 6
LITEMATIC_SUBVERSION = 1
SPONGE_VERSION = 2  # Sponge format version used by WorldEdit (https://github.com/SpongePowered/Schematic-Specification)
MC_DATA_VERSION = 2975  # Minecraft 1.18.2 (https://minecraft.wiki/w/Data_version)
DEFAULT_NAME = "Unnamed"  # Default name given to schematics and regions if unspecified
LITEMAPY_NAME = "Litemapy"  # Used to identify schematic created with Litemapy in metadata
LITEMAPY_VERSION = "0.10.0b0"  # Unique version string
